@@ -25,7 +25,7 @@
 
 extern const char *minor_str[];
 
-#define VERSION "v8.6"
+#define VERSION "v9.3"
 
 #ifndef O_DIRECT
 #define O_DIRECT	040000	/* direct disk access, not easily obtained from headers */
@@ -73,27 +73,26 @@ static int set_defects  = 0, get_defects  = 0, defects  = 0;
 static int set_wcache   = 0, get_wcache   = 0, wcache   = 0;
 static int set_doorlock = 0, get_doorlock = 0, doorlock = 0;
 static int set_seagate  = 0, get_seagate  = 0;
+static int get_idleimmediate = 0, set_idleimmediate = 0;
+static int get_idleunload = 0, set_idleunload = 0;
 static int set_standbynow = 0, get_standbynow = 0;
 static int set_sleepnow   = 0, get_sleepnow   = 0;
 static int set_powerup_in_standby = 0, get_powerup_in_standby = 0, powerup_in_standby = 0;
 static int get_hitachi_temp = 0, set_hitachi_temp = 0;
-static int set_freeze   = 0;
+static int security_freeze   = 0;
 static int security_master = 1, security_mode = 0;
 static int enhanced_erase = 0;
 static int set_security   = 0;
+static int do_dco_freeze = 0, do_dco_restore = 0, do_dco_identify = 0;
 static unsigned int security_command = ATA_OP_SECURITY_UNLOCK;
 
-static char security_password[33];
+static char security_password[33], *fwpath;
 
 static int get_powermode  = 0, set_powermode = 0;
 static int set_apmmode = 0, get_apmmode= 0, apmmode = 0;
 static int get_cdromspeed = 0, set_cdromspeed = 0, cdromspeed = 0;
 static int do_IDentity = 0, drq_hsm_error = 0;
-static int get_unregister = 0, set_unregister = 0, unregister = 0;
-static int	scan_hwif = 0;
-static int	hwif_data = 0;
-static int	hwif_ctrl = 0;
-static int	hwif_irq = 0;
+static int do_fwdownload = 0, final_80h = 0;
 static int	set_busstate = 0, get_busstate = 0, busstate = 0;
 static int	set_reread_partn = 0, get_reread_partn;
 static int	set_acoustic = 0, get_acoustic = 0, acoustic = 0;
@@ -111,8 +110,8 @@ static int   set_max_sectors = 0, set_max_permanent, get_native_max_sectors = 0;
 static __u64 set_max_addr = 0;
 
 static int	get_doreset = 0, set_doreset = 0;
-static int	get_tristate = 0, set_tristate = 0, tristate = 0;
 static int	i_know_what_i_am_doing = 0;
+static int	please_destroy_my_drive = 0;
 
 const int timeout_12secs = 12;
 const int timeout_2hrs   = (2 * 60 * 60);
@@ -138,7 +137,7 @@ const char *cfg_str[] =
 };
 
 const char *SlowMedFast[]	= {"slow", "medium", "fast", "eide", "ata"};
-const char *BuffType[]	= {"unknown", "1Sect", "DualPort", "DualPortCache"};
+const char *BuffType[4]		= {"unknown", "1Sect", "DualPort", "DualPortCache"};
 
 #define YN(b)	(((b)==0)?"no":"yes")
 
@@ -287,26 +286,6 @@ quit:
 	munmap(buf, TIMING_BUF_BYTES);
 }
 
-static int do_blkgetsize (int fd, unsigned long long *blksize64)
-{
-	int		rc;
-	unsigned int	blksize32 = 0;
-
-#ifdef BLKGETSIZE64
-	if (0 == ioctl(fd, BLKGETSIZE64, blksize64)) {	// returns bytes
-		*blksize64 /= 512;
-		return 0;
-	}
-#endif
-	rc = ioctl(fd, BLKGETSIZE, &blksize32);	// returns sectors
-	if (rc) {
-		rc = errno;
-		perror(" BLKGETSIZE failed");
-	}
-	*blksize64 = blksize32;
-	return rc;
-}
-
 static int time_device (int fd)
 {
 	char *buf;
@@ -319,11 +298,11 @@ static int time_device (int fd)
 	 * get device size
 	 */
 	if (do_ctimings || do_timings) {
-		unsigned long long blksize;
+		__u64 nsectors;
 		do_flush = 1;
-		err = do_blkgetsize(fd, &blksize);
+		err = get_dev_geometry(fd, NULL, NULL, NULL, NULL, &nsectors);
 		if (!err)
-			max_iterations = blksize / (2 * 1024) / TIMING_BUF_MB;
+			max_iterations = nsectors / (2 * 1024) / TIMING_BUF_MB;
 	}
 	buf = prepare_timing_buf(TIMING_BUF_BYTES);
 	if (!buf)
@@ -555,8 +534,9 @@ static void interpret_standby (void)
 				if (hrs)	  printf("%u hours", hrs);
 				if (hrs && mins)  printf(" + ");
 				if (mins)	  printf("%u minutes", mins);
-			} else
+			} else {
 				printf("illegal value");
+			}
 			break;
 	}
 	printf(")\n");
@@ -827,6 +807,16 @@ static void confirm_i_know_what_i_am_doing (const char *opt, const char *explana
 	}
 }
 
+static void confirm_please_destroy_my_drive (const char *opt, const char *explanation)
+{
+	if (!please_destroy_my_drive) {
+		fprintf(stderr, "Use of %s is EXTREMELY DANGEROUS.\n%s\n"
+		"Please also supply the --please-destroy-my-drive flag if you really want this.\n"
+		"Program aborted.\n", opt, explanation);
+		exit(EPERM);
+	}
+}
+
 static int flush_wcache (int fd, __u16 **id_p)
 {
 	__u8 args[4] = {ATA_OP_FLUSHCACHE,0,0,0};
@@ -843,60 +833,70 @@ static int flush_wcache (int fd, __u16 **id_p)
 	return err;
 }
 
-static void dump_sector (__u16 *w)
+static void dump_sectors (__u16 *w, unsigned int count)
 {
 	unsigned int i;
 
-	for (i = 0; i < (256/8); ++i) {
+	for (i = 0; i < (count*256/8); ++i) {
 		printf("%04x %04x %04x %04x %04x %04x %04x %04x\n", w[0], w[1], w[2], w[3], w[4], w[5], w[6], w[7]);
 		w += 8;
 	}
 }
 
-static __u64 get_start_lba (int fd)
+static int abort_if_not_full_device (int fd, __u64 lba, const char *devname, const char *msg)
 {
-	__u64 lba;
-
-	static struct local_hd_geometry      g;
-	static struct local_hd_big_geometry bg;
-
-	if (!ioctl(fd, HDIO_GETGEO_BIG, &bg)) {
-		lba = bg.start;
-	} else if (!ioctl(fd, HDIO_GETGEO, &g)) {
-		lba = g.start;
-	} else {
-		int err = errno;
-		perror(" HDIO_GETGEO failed: cannot determine correct LBA offset, aborting.");
-		exit(err);
-	}
-	return lba;
-}
-
-static int abort_if_not_full_device (int fd, __u64 lba, const char *devname)
-{
-	__u64 start = get_start_lba(fd);
-
+	__u64 start_lba;
+	int i, err, shortened = 0;
 	char *fdevname = strdup(devname);
-	int i, shortened = 0;
 
+	err = get_dev_geometry(fd, NULL, NULL, NULL, &start_lba, NULL);
+	if (err)
+		exit(err);
 	for (i = strlen(fdevname); --i > 2 && (fdevname[i] >= '0' && fdevname[i] <= '9');) {
 		fdevname[i] = '\0';
 		shortened = 1;
 	}
 
 	if (!shortened)
-		fdevname = "the full disk";
+		fdevname = strdup("the full disk");
 
-	if (start == 0ULL)
+	if (start_lba == 0ULL)
 		return 0;
-	fprintf(stderr, "Device %s has non-zero LBA starting offset of %llu.\n", devname, start);
-	fprintf(stderr, "Please use an absolute LBA with the /dev/ entry for the full device, rather than a partition name.\n");
-	fprintf(stderr, "%s is probably a partition of %s (?)\n", devname, fdevname);
-	fprintf(stderr, "The absolute LBA of sector %llu from %s should be %llu\n", lba, devname, start + lba);
+	if (msg) {
+		fprintf(stderr, "%s\n", msg);
+	} else {
+		fprintf(stderr, "Device %s has non-zero LBA starting offset of %llu.\n", devname, start_lba);
+		fprintf(stderr, "Please use an absolute LBA with the /dev/ entry for the full device, rather than a partition name.\n");
+		fprintf(stderr, "%s is probably a partition of %s (?)\n", devname, fdevname);
+		fprintf(stderr, "The absolute LBA of sector %llu from %s should be %llu\n", lba, devname, start_lba + lba);
+	}
 	fprintf(stderr, "Aborting.\n");
 	exit(EINVAL);
 }
 
+static __u16 *get_dco_identify_data (int fd, int quietly)
+{
+	static __u8 args[4+512];
+	__u16 *dco = (void *)(args + 4);
+	int i;
+	
+	memset(args, 0, sizeof(args));
+	args[0] = ATA_OP_DCO;
+	args[2] = 0xc2;
+	args[3] = 1;
+	if (do_drive_cmd(fd, args)) {
+		if (!quietly)
+			perror(" HDIO_DRIVE_CMD(dco_identify) failed");
+		return NULL;
+	} else {
+		/* byte-swap the little-endian DCO data to match byte-order on host CPU */
+		for (i = 0; i < 0x100; ++i)
+			__le16_to_cpus(&dco[i]);
+		//dump_sectors(dco, 1);
+		return dco;
+	}
+}
+
 static __u64 do_get_native_max_sectors (int fd, __u16 *id)
 {
 	int err = 0;
@@ -923,6 +923,9 @@ static __u64 do_get_native_max_sectors (int fd, __u16 *id)
 			err = errno;
 			perror (" READ_NATIVE_MAX_ADDRESS_EXT failed");
 		} else {
+			if (verbose)
+				printf("READ_NATIVE_MAX_ADDRESS_EXT response: hob={%02x %02x %02x} lob={%02x %02x %02x}\n",
+					r.hob.lbah, r.hob.lbam, r.hob.lbal, r.lob.lbah, r.lob.lbam, r.lob.lbal);
 			max = (((__u64)((r.hob.lbah << 16) | (r.hob.lbam << 8) | r.hob.lbal) << 24)
 				     | ((r.lob.lbah << 16) | (r.lob.lbam << 8) | r.lob.lbal)) + 1;
 		}
@@ -946,7 +949,7 @@ static int do_make_bad_sector (int fd, __u16 *id, __u64 lba, const char *devname
 	struct hdio_taskfile *r;
 	const char *flagged;
 
-	abort_if_not_full_device(fd, lba, devname);
+	abort_if_not_full_device(fd, lba, devname, NULL);
 	r = malloc(sizeof(struct hdio_taskfile) + 520);
 	if (!r) {
 		err = errno;
@@ -993,7 +996,7 @@ static int do_write_sector (int fd, __u64 lba, const char *devname)
 	__u8 ata_op;
 	struct hdio_taskfile *r;
 
-	abort_if_not_full_device(fd, lba, devname);
+	abort_if_not_full_device(fd, lba, devname, NULL);
 	r = malloc(sizeof(struct hdio_taskfile) + 512);
 	if (!r) {
 		err = errno;
@@ -1026,7 +1029,7 @@ static int do_read_sector (int fd, __u64 lba, const char *devname)
 	__u8 ata_op;
 	struct hdio_taskfile *r;
 
-	abort_if_not_full_device(fd, lba, devname);
+	abort_if_not_full_device(fd, lba, devname, NULL);
 	r = malloc(sizeof(struct hdio_taskfile) + 512);
 	if (!r) {
 		err = errno;
@@ -1044,12 +1047,61 @@ static int do_read_sector (int fd, __u64 lba, const char *devname)
 		perror("FAILED");
 	} else {
 		printf("succeeded\n");
-		dump_sector(r->data);
+		dump_sectors(r->data, 1);
 	}
 	free(r);
 	return err;
 }
 
+static int do_idleunload (int fd, const char *devname)
+{
+	int err = 0;
+	struct hdio_taskfile r;
+
+	abort_if_not_full_device(fd, 0, devname, NULL);
+	init_hdio_taskfile(&r, ATA_OP_IDLEIMMEDIATE, RW_READ, LBA28_OK, 0x0554e4c, 0, 0);
+	r.oflags.b.feat = 1;
+	r.lob.feat = 0x44;
+
+	if (do_taskfile_cmd(fd, &r, timeout_12secs)) {
+		err = errno;
+		perror("TASKFILE(idle_immediate_unload) failed");
+	}
+	return err;
+}
+
+#if 0
+static int do_read_sectors (int fd, __u64 lba, const char *devname)
+{
+	int err = 0;
+	__u8 ata_op;
+	struct hdio_taskfile *r;
+
+	abort_if_not_full_device(fd, lba, devname, NULL);
+	r = malloc(sizeof(struct hdio_taskfile) + (4*512));
+	if (!r) {
+		err = errno;
+		perror("malloc()");
+		return err;
+	}
+	ata_op = (lba >> 28) ? ATA_OP_READ_PIO_EXT : ATA_OP_READ_PIO;
+	init_hdio_taskfile(r, ata_op, RW_READ, LBA28_OK, lba, 4, 4*512);
+
+	printf("reading sectors %llu: ", lba);
+	fflush(stdout);
+
+	if (do_taskfile_cmd(fd, r, timeout_12secs)) {
+		err = errno;
+		perror("FAILED");
+	} else {
+		printf("succeeded\n");
+		dump_sectors(r->data, 4);
+	}
+	free(r);
+	return err;
+}
+#endif
+
 static int do_set_max_sectors (int fd, __u16 *id, __u64 max_lba, int permanent)
 {
 	int err = 0;
@@ -1085,7 +1137,7 @@ void process_dev (char *devname)
 	fd = open (devname, open_flags);
 	if (fd < 0) {
 		if (errno == EROFS) {
-			open_flags &= ~O_WRONLY;
+			open_flags = (open_flags & ~O_RDWR) | O_RDONLY;
 			fd = open (devname, open_flags);
 			if (fd >= 0)
 				goto open_ok;
@@ -1106,25 +1158,6 @@ open_ok:
 			perror(" BLKRASET failed");
 		}
 	}
-	if (set_unregister) {
-		if (get_unregister)
-			printf(" attempting to unregister hwif#%u\n", set_unregister);
-		if (ioctl(fd, HDIO_UNREGISTER_HWIF, set_unregister)) {
-			err = errno;
-			perror(" HDIO_UNREGISTER_HWIF failed");
-		}
-	}
-	if (scan_hwif) {
-		int	args[3];
-		printf(" attempting to scan hwif (0x%x, 0x%x, %u)\n", hwif_data, hwif_ctrl, hwif_irq);
-		args[0] = hwif_data;
-		args[1] = hwif_ctrl;
-		args[2] = hwif_irq;
-		if (ioctl(fd, HDIO_SCAN_HWIF, args)) {
-			err = errno;
-			perror(" HDIO_SCAN_HWIF failed");
-		}
-	}
 	if (set_piomode) {
 		if (get_piomode) {
 			if (piomode == 255)
@@ -1188,14 +1221,9 @@ open_ok:
 		}
 	}
 	if (set_dma_q) {
-		if (get_dma_q) {
-			printf(" setting DMA queue_depth to %d", dma_q);
-			on_off(dma_q);
-		}
-		if (ioctl(fd, HDIO_SET_QDMA, dma_q)) {
-			err = errno;
-			perror(" HDIO_SET_QDMA failed");
-		}
+		if (get_dma_q)
+			printf(" setting queue_depth to %d\n", dma_q);
+		err = sysfs_set_attr(fd, "device/queue_depth", "%u", &dma_q, 1);
 	}
 	if (set_nowerr) {
 		if (get_nowerr) {
@@ -1333,7 +1361,7 @@ open_ok:
 		/* FIXME: the CDROM_SELECT_SPEED ioctl
 		 * actually issues GPCMD_SET_SPEED to the drive.
 		 * But many newer DVD drives want GPCMD_SET_STREAMING instead.
-		 * Perhaps we should just do an sg16 issue here of *both* opcodes?
+		 * Perhaps we should just do an sg16 issue here of either opcode?
 		 */
 		if (get_cdromspeed)
 			printf ("setting cdrom speed to %d\n", cdromspeed);
@@ -1384,6 +1412,20 @@ open_ok:
 			perror(" HDIO_DRIVE_CMD(standby) failed");
 		}
 	}
+	if (set_idleimmediate) {
+		__u8 args[4] = {ATA_OP_IDLEIMMEDIATE,0,0,0};
+		if (get_idleimmediate)
+			printf(" issuing idle_immediate command\n");
+		if (do_drive_cmd(fd, args)) {
+			err = errno;
+			perror(" HDIO_DRIVE_CMD(idle_immediate) failed");
+		}
+	}
+	if (set_idleunload) {
+		if (get_idleunload)
+			printf(" issuing idle_immediate_unload command\n");
+		err = do_idleunload(fd, devname);
+	}
 	if (set_sleepnow) {
 		__u8 args1[4] = {ATA_OP_SLEEPNOW1,0,0,0};
 		__u8 args2[4] = {ATA_OP_SLEEPNOW2,0,0,0};
@@ -1398,12 +1440,34 @@ open_ok:
 	if (set_security) {
 		do_set_security(fd);
 	}
-	if (set_freeze) {
-		__u8 args[4] = {ATA_OP_SECURITY_FREEZE_LOCK,0,0,0};
-		printf(" issuing Security Freeze command\n");
+	if (do_dco_identify) {
+		__u16 *dco = get_dco_identify_data(fd, 0);
+		if (dco)
+			dco_identify_print(dco);
+	}
+	if (do_dco_restore) {
+		__u8 args[4] = {ATA_OP_DCO,0,0xc0,0};
+		confirm_i_know_what_i_am_doing("--dco-restore", "You are trying to deliberately reset your drive configuration back to the factory defaults.  This may change the apparent capacity and feature set of the drive, making all data on it inaccessible.  You could lose *everything*.");
+		printf(" issuing DCO restore command\n");
 		if (do_drive_cmd(fd, args)) {
 			err = errno;
-			perror(" HDIO_DRIVE_CMD(freeze) failed");
+			perror(" HDIO_DRIVE_CMD(dco_restore) failed");
+		}
+	}
+	if (do_dco_freeze) {
+		__u8 args[4] = {ATA_OP_DCO,0,0xc1,0};
+		printf(" issuing DCO freeze command\n");
+		if (do_drive_cmd(fd, args)) {
+			err = errno;
+			perror(" HDIO_DRIVE_CMD(dco_freeze) failed");
+		}
+	}
+	if (security_freeze) {
+		__u8 args[4] = {ATA_OP_SECURITY_FREEZE_LOCK,0,0,0};
+		printf(" issuing security freeze command\n");
+		if (do_drive_cmd(fd, args)) {
+			err = errno;
+			perror(" HDIO_DRIVE_CMD(security_freeze) failed");
 		}
 	}
 	if (set_seagate) {
@@ -1416,14 +1480,14 @@ open_ok:
 		}
 	}
 	if (set_standby) {
-		__u8 args[4] = {ATA_OP_SETIDLE1,standby,0,0};
+		__u8 args[4] = {ATA_OP_SETIDLE,standby,0,0};
 		if (get_standby) {
 			printf(" setting standby to %u", standby);
 			interpret_standby();
 		}
 		if (do_drive_cmd(fd, args)) {
 			err = errno;
-			perror(" HDIO_DRIVE_CMD(setidle1) failed");
+			perror(" HDIO_DRIVE_CMD(setidle) failed");
 		}
 	}
 	if (set_busstate) {
@@ -1452,6 +1516,15 @@ open_ok:
 		confirm_i_know_what_i_am_doing("--write-sector", "You are trying to deliberately overwrite a low-level sector on the media\nThis is a BAD idea, and can easily result in total data loss.");
 		err = do_write_sector(fd, write_sector_addr, devname);
 	}
+	if (do_fwdownload) {
+		abort_if_not_full_device (fd, 0, devname, "--fwdownload requires the raw device, not a partition.");
+		confirm_i_know_what_i_am_doing("--fwdownload", "You are trying to deliberately overwrite the drive firmware with the contents of the specified file.  If this fails, your drive could be toast.");
+		confirm_please_destroy_my_drive("--fwdownload", "This might destroy the drive and/or all data on it.");
+		id = get_identify_data(fd, id);
+		err = fwdownload(fd, id, fwpath, final_80h);
+		if (err)
+			exit(err);
+	}
 	if (read_sector) {
 		err = do_read_sector(fd, read_sector_addr, devname);
 	}
@@ -1475,8 +1548,7 @@ open_ok:
 		if (do_drive_cmd(fd, args)) {
 			err = errno;
 			perror(" HDIO_DRIVE_CMD(hitachisensecondition) failed");
-		}
-		else {
+		} else {
 			printf(" drive temperature (celsius) is:  ");
 			if (args[2]==0)
 				printf("under -20");
@@ -1502,6 +1574,7 @@ open_ok:
 			printf(" IO_support    =%3ld (", parm);
 			switch (parm) {
 				case 0:	printf("default) \n");
+					break;
 				case 2: printf("16-bit)\n");
 					break;
 				case 1:	printf("32-bit)\n");
@@ -1512,7 +1585,7 @@ open_ok:
 					break;
 				default:printf("\?\?\?)\n");
 			}
-               } else {
+               } else if (get_io32bit) {
                        err = errno;
                        perror(" HDIO_GET_32BIT failed");
 		}
@@ -1521,7 +1594,7 @@ open_ok:
 		if (0 == ioctl(fd, HDIO_GET_UNMASKINTR, &parm)) {
 			printf(" unmaskirq     = %2ld", parm);
 			on_off(parm);
-               } else {
+               } else if (get_unmask) {
                        err = errno;
                        perror(" HDIO_GET_UNMASKINTR failed");
 		}
@@ -1534,25 +1607,21 @@ open_ok:
 				printf(" (DMA-Assisted-PIO)\n");
 			else
 				on_off(parm);
-                } else {
+                } else if (get_dma) {
                        err = errno;
                        perror(" HDIO_GET_DMA failed");
 		}
 	}
 	if (get_dma_q) {
-		if(ioctl(fd, HDIO_GET_QDMA, &parm)) {
-			err = errno;
-			perror(" HDIO_GET_QDMA failed");
-		} else {
-			printf(" queue_depth   = %2ld", parm);
-			on_off(parm);
-		}
+		err = sysfs_get_attr(fd, "device/queue_depth", "%u", &dma_q, NULL, 1);
+		if (!err)
+			printf(" queue_depth   = %2u\n", dma_q);
 	}
 	if (do_defaults || get_keep) {
 		if (0 == ioctl(fd, HDIO_GET_KEEPSETTINGS, &parm)) {
 			printf(" keepsettings  = %2ld", parm);
 			on_off(parm);
-		} else {
+		} else if (get_keep) {
 			err = errno;
                         perror(" HDIO_GET_KEEPSETTINGS failed");
 		}
@@ -1585,25 +1654,13 @@ open_ok:
 		}
 	}
 	if (do_defaults || get_geom) {
-		unsigned long long blksize;
-		static const char msg[] = " geometry      = %u/%u/%u, sectors = %lld, start = %ld\n";
-/*
- * Note to self:  when compiled 32-bit (AMD,Mips64), the userspace version of this struct
- * is going to be 32-bits smaller than the kernel representation.. random stack corruption!
- */
-		static struct local_hd_geometry      g;
-		static struct local_hd_big_geometry bg;
+		__u32 cyls = 0, heads = 0, sects = 0;
+		__u64 start_lba = 0, nsectors = 0;
+		err = get_dev_geometry (fd, &cyls, &heads, &sects, &start_lba, &nsectors);
 
-		err = do_blkgetsize(fd, &blksize);
-		if (!err) {
-			if (!ioctl(fd, HDIO_GETGEO_BIG, &bg))
-				printf(msg, bg.cylinders, bg.heads, bg.sectors, blksize, bg.start);
-			else if (ioctl(fd, HDIO_GETGEO, &g)) {
-				err = errno;
-				perror(" HDIO_GETGEO failed");
-			} else
-				printf(msg, g.cylinders, g.heads, g.sectors, blksize, g.start);
-		}
+		if (!err)
+			printf(" geometry      = %u/%u/%u, sectors = %lld, start = %lld\n",
+				cyls, heads, sects, nsectors, start_lba);
 	}
 	if (get_powermode) {
 		__u8 args[4] = {ATA_OP_CHECKPOWERMODE1,0,0,0};
@@ -1613,8 +1670,9 @@ open_ok:
 		 && do_drive_cmd(fd, args)) {
 			err = errno;
 			state = "unknown";
-		} else
+		} else {
 			state = (args[2] == 255) ? "active/idle" : "standby";
+		}
 		printf(" drive state is:  %s\n", state);
 	}
 	if (do_identity) {
@@ -1623,12 +1681,13 @@ open_ok:
 		if (!ioctl(fd, HDIO_GET_IDENTITY, id2)) {
 			if (multcount != -1) {
 				id2[59] = multcount | 0x100;
-			} else
+			} else {
 				id2[59] &= ~0x100;
+			}
 			dump_identity(id2);
-		} else if (errno == -ENOMSG)
+		} else if (errno == -ENOMSG) {
 			printf(" no identification info available\n");
-		else {
+		} else {
 			err = errno;
 			perror(" HDIO_GET_IDENTITY failed");
 		}
@@ -1637,7 +1696,7 @@ open_ok:
 		id = get_identify_data(fd, id);
 		if (id) {
 			if (do_IDentity == 2)
-				dump_sector(id);
+				dump_sectors(id, 1);
 			else
 				identify((void *)id);
 		}
@@ -1682,8 +1741,9 @@ open_ok:
 		if (ioctl(fd, HDIO_GET_BUSSTATE, &parm)) {
 			err = errno;
 			perror(" HDIO_GET_BUSSTATE failed");
-		} else
+		} else {
 			printf(" busstate      = %2ld (%s)\n", parm, busstate_str(parm));
+		}
 	}
 	if (get_native_max_sectors) {
 		__u64 visible, native;
@@ -1699,8 +1759,18 @@ open_ok:
 					printf(", HPA is enabled\n");
 				else if (visible == native)
 					printf(", HPA is disabled\n");
-				else
-					printf(", HPA setting seems invalid\n");
+				else {
+					__u16 *dco = get_dco_identify_data(fd, 1);
+					if (dco) {
+						__u64 dco_max = dco[5];
+						dco_max = ((((__u64)dco[5]) << 32) | (dco[4] << 16) | dco[3]) + 1;
+						printf("(%llu?)", dco_max);
+					}
+					printf(", HPA setting seems invalid");
+					if ((native & 0xffffff000000ull) == 0)
+						printf(" (buggy kernel device driver?)");
+					putchar('\n');
+				}
 			}
 		}
 	}
@@ -1729,15 +1799,6 @@ open_ok:
 			perror(" HDIO_DRIVE_RESET failed");
 		}
 	}
-	if (set_tristate) {
-		__u8 args[4] = {0,tristate,0,0};
-		if (get_tristate)
-			printf(" setting tri-state to %d\n", tristate);
-		if (ioctl(fd, HDIO_TRISTATE_HWIF, &args)) {
-			err = errno;
-			perror(" HDIO_TRISTATE_HWIF failed");
-		}
-	}
 	close (fd);
 	if (err)
 		exit (err);
@@ -1776,27 +1837,35 @@ static void usage_help (int rc)
 	" -p   set PIO mode on IDE interface chipset (0,1,2,3,4,...)\n"
 	" -P   set drive prefetch count\n"
 	" -q   change next setting quietly\n"
-	" -Q   get/set DMA tagged-queuing depth (if supported)\n"
+	" -Q   get/set DMA queue_depth (if supported)\n"
 	" -r   get/set device  readonly flag (DANGEROUS to set)\n"
-	" -R   register an IDE interface (DANGEROUS)\n"
+	" -R   obsolete\n"
 	" -s   set power-up in standby flag (0/1) (DANGEROUS)\n"
 	" -S   set standby (spindown) timeout\n"
 	" -t   perform device read timings\n"
 	" -T   perform cache read timings\n"
 	" -u   get/set unmaskirq flag (0/1)\n"
-	" -U   un-register an IDE interface (DANGEROUS)\n"
+	" -U   obsolete\n"
 	" -v   defaults; same as -acdgkmur for IDE drives\n"
 	" -V   display program version and exit immediately\n"
 	" -w   perform device reset (DANGEROUS)\n"
 	" -W   get/set drive write-caching flag (0/1)\n"
-	" -x   tristate device for hotswap (0/1) (DANGEROUS)\n"
+	" -x   obsolete\n"
 	" -X   set IDE xfer mode (DANGEROUS)\n"
 	" -y   put drive in standby mode\n"
 	" -Y   put drive to sleep\n"
 	" -Z   disable Seagate auto-powersaving mode\n"
 	" -z   re-read partition table\n"
+	" --dco-freeze      freeze/lock current device configuration until next power cycle\n"
+	" --dco-identify    read/dump device configuration identify data\n"
+	" --dco-restore     reset device configuration back to factory defaults\n"
 	" --direct          use O_DIRECT to bypass page cache for timings\n"
 	" --drq-hsm-error   crash system with a \"stuck DRQ\" error (VERY DANGEROUS)\n"
+	" --fibmap          show device extents (and fragmentation) for a file\n"
+	" --fibmap-sector   show absolute LBA of a specfic sector of a file\n"
+	" --fwdownload      Download firmware file to drive (EXTREMELY DANGEROUS)\n"
+	" --idle-immediate  idle drive immediately\n"
+	" --idle-unload     idle immediately and unload heads\n"
 	" --Istdin          read identify data from stdin as ASCII hex\n"
 	" --Istdout         write identify data to stdout as ASCII hex\n"
 	" --make-bad-sector deliberately corrupt a sector directly on the media (VERY DANGEROUS)\n"
@@ -1931,7 +2000,6 @@ numeric_parm (char c, const char *name, int *val, int *setparm, int *getparm, in
 		got_digit = 1;
 	}
 	if ((set_only && !got_digit) || *val < min || *val > max) {
-		*val = min;	//FIXME?
 		fprintf(stderr, "  -%c: bad/missing %s value (%d..%d)\n", c, name, min, max);
 		exit(EINVAL);
 	}
@@ -1940,7 +2008,8 @@ numeric_parm (char c, const char *name, int *val, int *setparm, int *getparm, in
 #define NUMERIC_PARM(CH,NAME,VAR,MIN,MAX,GETSET) numeric_parm(CH,NAME,&VAR,&set_##VAR,&get_##VAR,MIN,MAX,GETSET)
 #define GET_SET_PARM(CH,NAME,VAR,MIN,MAX) CH:NUMERIC_PARM(CH,NAME,VAR,MIN,MAX,0);break
 #define     SET_PARM(CH,NAME,VAR,MIN,MAX) CH:NUMERIC_PARM(CH,NAME,VAR,MIN,MAX,1);break
-#define     SET_FLAG(CH,VAR)              CH:get_##VAR=noisy;noisy=1;set_##VAR=1;break
+#define     SET_FLAG1(VAR)                get_##VAR=noisy;noisy=1;set_##VAR=1
+#define     SET_FLAG(CH,VAR)              CH:SET_FLAG1(VAR);break
 #define      DO_FLAG(CH,VAR)              CH:VAR=1;noisy=1;break
 #define    INCR_FLAG(CH,VAR)              CH:VAR++;noisy=1;break
 
@@ -1970,7 +2039,7 @@ static void get_security_password (int handle_NULL)
 }
 
 static int
-get_sector_value (int optional, const char flag_c, int *flag_p, __u64 *value_p, unsigned int min_value, const char *emsg)
+get_lba_parm (int optional, const char flag_c, int *flag_p, __u64 *value_p, unsigned int min_value, const char *emsg)
 {
 	int got_value = 0, got_digit = 0;
 	const __u64 limit = (1ULL << 48) - 1;
@@ -2012,7 +2081,7 @@ get_set_max_sectors_parms (void)
 {
 	get_native_max_sectors = noisy;
 	noisy = 1;
-	set_max_sectors = get_sector_value(1, 'p', &set_max_permanent, &set_max_addr, 1, "-N");
+	set_max_sectors = get_lba_parm(1, 'p', &set_max_permanent, &set_max_addr, 1, "-N");
 }
 
 static void
@@ -2033,7 +2102,9 @@ handle_standalone_longarg (char *name)
 		identify_from_stdin();
 		exit(0);
 	}
-	if (0 == strcasecmp(name, "security-help")) {
+	if (0 == strcasecmp(name, "dco-restore")) {
+		do_dco_restore = 1;
+	} else if (0 == strcasecmp(name, "security-help")) {
 		security_help(0);
 		exit(0);
 	} else if (0 == strcasecmp(name, "security-unlock")) {
@@ -2062,6 +2133,50 @@ handle_standalone_longarg (char *name)
 	}
 }
 
+static void
+get_filename_parm (char **result, const char *emsg)
+{
+	if (!*argp && argc)
+		argp = *argv++, --argc;
+	if (!argp || !*argp) {
+		fprintf(stderr, "  %s: bad/missing filename parameter\n", emsg);
+		exit(EINVAL);
+	}
+	*result = argp;
+	argp += strlen(argp);
+	// if (argc) argp = *argv++, --argc;
+}
+
+static void
+do_fibmap_sector (const char *name)
+{
+	int err;
+	char *path;
+	__u64 sector, lba;
+
+	get_filename_parm(&path, name);
+	get_lba_parm(0, 0, NULL, &sector, 0, name);
+	if (num_flags_processed || argc)
+		usage_help(EINVAL);
+	err = do_fibmap(path, sector, &lba);
+	if (!err)
+		printf("%s[%llu]: %llu\n", path, sector, lba);
+	else if (err == EBADSLT)
+		fprintf(stderr, "%s[%llu]: unallocated\n", path, sector);
+	exit(err);
+}
+
+static void
+do_fibmap_file (const char *name)
+{
+	char *path;
+
+	get_filename_parm(&path, name);
+	if (num_flags_processed || argc)
+		usage_help(EINVAL);
+	exit(do_fibmap(path, 0, NULL));
+}
+
 static int
 get_longarg (void)
 {
@@ -2075,20 +2190,42 @@ get_longarg (void)
 	} else if (0 == strcasecmp(name, "yes-i-know-what-i-am-doing")) {
 		i_know_what_i_am_doing = 1;
 		--num_flags_processed;	/* doesn't count as an action flag */
+	} else if (0 == strcasecmp(name, "please-destroy-my-drive")) {
+		please_destroy_my_drive = 1;
+		--num_flags_processed;	/* doesn't count as an action flag */
 	} else if (0 == strcasecmp(name, "direct")) {
 		open_flags |= O_DIRECT;
 		--num_flags_processed;	/* doesn't count as an action flag */
 	} else if (0 == strcasecmp(name, "drq-hsm-error")) {
 		drq_hsm_error = 1;
+	} else if (0 == strcasecmp(name, "dco-freeze")) {
+		do_dco_freeze = 1;
+	} else if (0 == strcasecmp(name, "dco-identify")) {
+		do_dco_identify = 1;
+	} else if (0 == strcasecmp(name, "fibmap-sector")) {
+		do_fibmap_sector(name);
+	} else if (0 == strcasecmp(name, "fibmap")) {
+		do_fibmap_file(name);
+	} else if (0 == strcasecmp(name, "fwdownload")) {
+		get_filename_parm(&fwpath, name);
+		do_fwdownload = 1;
+	} else if (0 == strcasecmp(name, "fwdownload80")) {
+		get_filename_parm(&fwpath, name);
+		do_fwdownload = 1;
+		final_80h = 1;
+	} else if (0 == strcasecmp(name, "idle-immediate")) {
+		SET_FLAG1(idleimmediate);
+	} else if (0 == strcasecmp(name, "idle-unload")) {
+		SET_FLAG1(idleunload);
 	} else if (0 == strcasecmp(name, "make-bad-sector")) {
 		make_bad_sector = 1;
-		get_sector_value(0, 'f', &make_bad_sector_flagged, &make_bad_sector_addr, 0, "--make-bad-sector");
+		get_lba_parm(0, 'f', &make_bad_sector_flagged, &make_bad_sector_addr, 0, name);
 	} else if (0 == strcasecmp(name, "write-sector") || 0 == strcasecmp(name, "repair-sector")) {
 		write_sector = 1;
-		get_sector_value(0, 0, NULL, &write_sector_addr, 0, "--write-sector");
+		get_lba_parm(0, 0, NULL, &write_sector_addr, 0, name);
 	} else if (0 == strcasecmp(name, "read-sector")) {
 		read_sector = 1;
-		get_sector_value(0, 0, NULL, &read_sector_addr, 0, "--read-sector");
+		get_lba_parm(0, 0, NULL, &read_sector_addr, 0, name);
 	} else if (0 == strcasecmp(name, "Istdout")) {
 		do_IDentity = 2;
 	} else if (0 == strcasecmp(name, "security-mode")) {
@@ -2116,7 +2253,7 @@ get_longarg (void)
 		}
 		--num_flags_processed;	/* doesn't count as an action flag */
 	} else if (0 == strcasecmp(name, "security-freeze")) {
-		set_freeze = 1;
+		security_freeze = 1;
 	} else {
 		handle_standalone_longarg(name);
 		return 1; /* 1 == no more flags allowed */
@@ -2185,19 +2322,17 @@ int main (int _argc, char **_argv)
 				case              'N': get_set_max_sectors_parms(); break;
 				case     SET_PARM('P',"prefetch",prefetch,0,255);
 				case              'q': quiet = 1; noisy = 0; break;
-				case     SET_PARM('Q',"queue-depth",dma_q,0,32);
+				case GET_SET_PARM('Q',"queue-depth",dma_q,0,1024);
 				case     SET_PARM('s',"powerup-in-standby",powerup_in_standby,0,1);
 				case     SET_PARM('S',"standby-interval",standby,0,255);
 				case GET_SET_PARM('r',"read-only",readonly,0,1);
 				case      DO_FLAG('t',do_timings);
 				case      DO_FLAG('T',do_ctimings);
 				case GET_SET_PARM('u',"unmask-irq",unmask,0,1);
-				case     SET_PARM('U',"unregister-interface",unregister,0,63);
 				case      DO_FLAG('v',do_defaults);
 				case              'V': fprintf(stdout, "%s %s\n", progname, VERSION); exit(0);
 				case     SET_FLAG('w',doreset);
 				case GET_SET_PARM('W',"write-cache",wcache,0,1);
-				case     SET_PARM('x',"tristate-device",tristate,0,1);
 				case     SET_FLAG('y',standbynow);
 				case     SET_FLAG('Y',sleepnow);
 				case     SET_FLAG('z',reread_partn);
@@ -2222,31 +2357,6 @@ int main (int _argc, char **_argv)
 						fprintf(stderr, "-X: missing value\n");
 					break;
 
-				case 'R':
-					if (!*argp && argc && isdigit(**argv))
-						argp = *argv++, --argc;
-					if(! argp) {
-						fprintf(stderr, "expected hwif_data\n");
-						exit(EINVAL);
-					}
-					sscanf(argp++, "%i", &hwif_data);
-					if (argc && isdigit(**argv))
-						argp = *argv++, --argc;
-					else {
-						fprintf(stderr, "expected hwif_ctrl\n");
-						exit(EINVAL);
-					}
-					sscanf(argp, "%i", &hwif_ctrl);
-					if (argc && isdigit(**argv))
-						argp = *argv++, --argc;
-					else {
-						fprintf(stderr, "expected hwif_irq\n");
-						exit(EINVAL);
-					}
-					sscanf(argp, "%i", &hwif_irq);
-					*argp = '\0';
-					scan_hwif = 1;
-					break;
 
 				default:
 					usage_help(EINVAL);
@@ -2256,5 +2366,5 @@ int main (int _argc, char **_argv)
 		if (!argc)
 			usage_help(EINVAL);
 	}
-	exit (0);
+	return 0;
 }
